#include <linux/fs.h>
#include <linux/msdos_fs.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include "dmsdos.h"

/* Here we do a virtual sector translation */

#define FAKED_BH_MAGIC 0xfff0
/* i hope this never occurs in kernel buffers... */

extern unsigned long int dmsdos_speedup;

/* cluster caching ... */
Cluster_head ccache[CCACHESIZE];

struct wait_queue * ccachewait=NULL;
int ccachelock=0;
void lock_ccache(void)
{ while(ccachelock)sleep_on(&ccachewait);
  ccachelock=1;
}
void unlock_ccache(void)
{ ccachelock=0;
  wake_up(&ccachewait);
}

struct wait_queue * fullwait=NULL;

int help_ccache=0;

/* must be called locked */
Cluster_head* find_in_ccache(struct super_block*sb,
                             int clusternr,Cluster_head**lastfree,
                             Cluster_head**oldest)
{ int i;
  int lastfree_n=-1;
  int oldest_n=-1;
  unsigned int oldest_time=CURRENT_TIME;
  
  for(i=0;i<CCACHESIZE;++i)
  { if(ccache[i].c_flags==C_FREE&&ccache[i].c_locked==0)lastfree_n=i;
                               /*  ^^^^^^^^^^^^^^^^^^^^ don't return locked
                                                        free clusters...*/
    else 
    { if(ccache[i].c_time<oldest_time&&ccache[i].c_count==0)
      { oldest_n=i;
        oldest_time=ccache[i].c_time;
      }
      if(ccache[i].c_clusternr==clusternr&&ccache[i].c_sb->s_dev==sb->s_dev)
      { /* found */
        return &(ccache[i]);
      }
    }
  }
  if(lastfree_n>=0&&lastfree!=NULL)*lastfree=&(ccache[lastfree_n]);
  /* uhhh, deadlock possible here... we should always have an oldest when
     there's nothing free any more... */
  /* use help_ccache to avoid always choosing the same cluster */
  if(lastfree_n<0&&oldest_n<0&&oldest!=NULL)
  { /* search more carefully */
    /* candidates are all those with count==0 */
    for(i=0;i<CCACHESIZE;++i)
    { ++help_ccache; if(help_ccache>=CCACHESIZE)help_ccache=0;
      if(ccache[help_ccache].c_count==0)
      { oldest_n=help_ccache;
        break;
      }
    }
  }
  if(oldest_n>=0&&oldest!=NULL)*oldest=&(ccache[oldest_n]);
  return NULL;
}

void unlock_ch(Cluster_head*ch)
{ ch->c_locked=0;
  wake_up(&(ch->c_wait));
}

Cluster_head* u_read_ch(struct super_block*sb,int clusternr, int doread)
{ Cluster_head*free=NULL;
  Cluster_head*oldest=NULL;
  Cluster_head*actual;
  int fail_trials=0;
  struct super_block*sb2;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
  
retry:
  lock_ccache();
  
  actual=find_in_ccache(sb,clusternr,&free,&oldest);
  
  if(actual)
  { /* if it is locked we must wait */
    if(actual->c_locked)
    { unlock_ccache();
      sleep_on(&(actual->c_wait));
      /* anything may have happened meanwhile */
      goto retry;
    }
    else actual->c_locked=1;
  }
  else if(free)
  { /* if it is locked we must wait */
    if(free->c_locked)
    { unlock_ccache();
      sleep_on(&(free->c_wait));
      /* anything may have happened meanwhile */
      goto retry;
    }
    else free->c_locked=1;
  }
  else if(oldest)
  { /* if it is locked we must wait */
    if(oldest->c_locked)
    { unlock_ccache();
      sleep_on(&(oldest->c_wait));
      /* anything may have happened meanwhile */
      goto retry;
    }
    else oldest->c_locked=1;
  }
  
  /* now the cluster to be used is locked, we don't need to lock the
     whole cache any more */
  unlock_ccache();
  
  if(actual==NULL)
  { if(free==NULL)
    { /* we must discard one */
      if(oldest==NULL)
      { /* oh, all are in use - we must wait */
        /*unlock_ccache();*/
        LOG_CCACHE("DMSDOS: cluster cache full, waiting...\n");
        sleep_on(&fullwait);
        goto retry;
      }
      /* we use oldest */
      /* write it before if it is dirty... */
      if(oldest->c_flags==C_DIRTY_NORMAL)
      { sb2=oldest->c_sb;
        if(dmsdos_write_cluster(sb2,oldest->c_data,oldest->c_length,
                             oldest->c_clusternr,0,0)<0)
          { write_failed:
            oldest->c_time=CURRENT_TIME;
            ++fail_trials;
            printk(KERN_EMERG "DMSDOS: cannot write dirty cluster %d, trying again later. This is %d. warning.\n",
                   oldest->c_clusternr, fail_trials);
            if(fail_trials<CCACHESIZE/5)
            { /*unlock_ccache();*/
              unlock_ch(oldest);
              goto retry;
            }
            unlock_ch(oldest);
            panic("DMSDOS: Too many write failures, giving up.\n");
          }
      }
      if(oldest->c_flags==C_DIRTY_UNCOMPR)
      { sb2=oldest->c_sb;
        if(dmsdos_write_cluster(sb2,oldest->c_data,oldest->c_length,
             oldest->c_clusternr,0,1)<0)goto write_failed;
      }
      FREE(oldest->c_data);
      oldest->c_flags=C_FREE;
      free=oldest;
      /* now we have a free one */
    }
    /* we have a free entry */
    free->c_data=MALLOC(dblsb->s_sectperclust*SECTOR_SIZE);
    if(free->c_data==NULL)
    { /* now we have a serious problem: no memory available */
      /* we cannot simply retry since this locks up the system 
         so we hope waiting for the next buffer to free up solves this
         but if there's no buffer any more to free up we hang forever...
         no better idea here... */
      printk("DMSDOS: ccache: no memory, waiting...\n");
      /*unlock_ccache();*/
      unlock_ch(free);
      sleep_on(&fullwait);
      goto retry;
    }
    free->c_count=0;
    free->c_flags=C_USED; /* it's no longer free */
    free->c_clusternr=clusternr;
    free->c_sb=sb;
    free->c_length=0;
    /* now we need the data from disk */
    if(doread)
    {
      free->c_length=dmsdos_read_cluster(sb,free->c_data,clusternr);
      if(free->c_length<0)
      { /* uh oh, read failed */
        printk("DMSDOS: ccache: read_cluster failed.\n");
        /* throw away */
        FREE(free->c_data);
        free->c_flags=C_FREE;
        /*unlock_ccache();*/
        unlock_ch(free);
        return NULL;
      }
    }
    /* now we have the actual data */
    actual=free;
  }
  
  /* setup time stamp and counter */
  actual->c_time=CURRENT_TIME;
  ++(actual->c_count);
  
  /*unlock_ccache();*/
  unlock_ch(actual);
  
  return actual;
}

/* this routine *must* return a cluster with its slack zerod out if the
   cluster doesn't have maximum length */
Cluster_head* ch_read(struct super_block*sb,int clusternr)
{ Cluster_head*ch;

  LOG_CCACHE("DMSDOS: ch_read cluster %d\n",clusternr);
  ch=u_read_ch(sb,clusternr,1);
  if(ch!=NULL){LOG_CCACHE("DMSDOS: ch_read: count=%d\n",ch->c_count);}
  LOG_CCACHE("DMSDOS: ch_read finished.\n");
  return ch;
}

/* this routine *needn't* zero out the data */
Cluster_head* ch_noread(struct super_block*sb,int clusternr)
{ Cluster_head*ch;

  LOG_CCACHE("DMSDOS: ch_noread cluster %d\n",clusternr);
  ch=u_read_ch(sb,clusternr,0);
  if(ch!=NULL){LOG_CCACHE("DMSDOS: ch_noread: count=%d\n",ch->c_count);}
  LOG_CCACHE("DMSDOS: ch_noread finished.\n");
  
  return ch;
}

int ch_dirty(Cluster_head*ch, int near, int ucflag)
{    /* Hmm...
        We must ensure that we can safely delay writing of the cluster
        before. This is done by calling the write_cluster function with
        the SIMULATE flag. If there's a problem then it is expected to
        return an error code (return value<0). In that case we fall back to
        the old write-through caching mechanism.
     */

  int i;
  struct super_block*sb;
  
  sb=ch->c_sb;

  LOG_CCACHE("DMSDOS: ch_dirty cluster %d flags %d ucflag %d\n",
                ch->c_clusternr,ch->c_flags,ucflag);

  /* if the cluster is locked we'd better wait for the lock going away */
  while(ch->c_locked)sleep_on(&(ch->c_wait));
  ch->c_locked=1;
       
  /* do not write free or deleted clusters */
  if(ch->c_flags==C_FREE||ch->c_flags==C_DELETED){unlock_ch(ch);return 0;}
     
  if(dmsdos_speedup&SP_USE_WRITE_BACK)
  {
     /* call write_cluster with SIMULATE flag (ucflag=2) first */
     i=dmsdos_write_cluster(sb,ch->c_data,ch->c_length,ch->c_clusternr,
                                 near,2);
     if(i>=0)
     { /* success - we can safely delay writing this cluster */
       ch->c_flags=ucflag>0?C_DIRTY_UNCOMPR:C_DIRTY_NORMAL;
       unlock_ch(ch);
       return i;
     }
     /* otherwise fall back to the default write-through code */
  }
     
  i=dmsdos_write_cluster(sb,ch->c_data,ch->c_length,ch->c_clusternr,
                                      near,ucflag);
  unlock_ch(ch);
  return i;
}

/* This routine is called if extremely important data (directory structure)
   really must be written when normal ch_dirty already fails. However, there's
   no guarantee that the data can really be written some time later. This
   function is meant as a last resort. Code will panic later if too much
   clusters are in the "dirty error wait queue". */
void ch_dirty_retry_until_success(Cluster_head*ch, int near, int ucflag)
{ if(ch->c_flags!=C_FREE&&ch->c_flags!=C_DELETED)
       ch->c_flags=ucflag>0?C_DIRTY_UNCOMPR:C_DIRTY_NORMAL;
}

void ch_free(Cluster_head*ch)
{ 
  --(ch->c_count);
  LOG_CCACHE("DMSDOS: ch_free cluster %d count_after_free %d\n",
             ch->c_clusternr,ch->c_count);
  if(ch->c_count==0)
  { /* throw away deleted clusters immediately */
    if(ch->c_flags==C_DELETED)
    { FREE(ch->c_data);
      ch->c_flags=C_FREE;
    } 
    wake_up(&fullwait);
  }
}

void ccache_init()
{ int i;

  for(i=0;i<CCACHESIZE;++i)
  { ccache[i].c_flags=C_FREE;
    ccache[i].c_time=0;
    ccache[i].c_count=0;
    ccache[i].c_wait=NULL;
    ccache[i].c_locked=0;
  }
}

/* for unmount */
/* we assume that the clusters of the device to be unmounted are
   - not in use (count==0) and
   - not locked (locked==0)
   either of them would mean that an inode of the filesystem is currently
   in use and thus the filesystem driver shouldn't allow unmount...
   ...violation will cause segfaults in a lot of other places...
*/
void free_ccache_dev(struct super_block*sb)
{ int i;

  lock_ccache();
  for(i=0;i<CCACHESIZE;++i)
  { if(ccache[i].c_flags!=C_FREE&&ccache[i].c_sb->s_dev==sb->s_dev)
    { 
      /* write it before if it is dirty... */
      if(ccache[i].c_flags==C_DIRTY_NORMAL)
      { /* we do not check for failure since we cannot do anything for it
           at unmount time (except panic, but this would be worse) */
        dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                             ccache[i].c_clusternr,0,UMOUNT_UCFLAG);
      }
      if(ccache[i].c_flags==C_DIRTY_UNCOMPR)
      { dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                             ccache[i].c_clusternr,0,1);
      }

      if(ccache[i].c_count)printk("DMSDOS: free_ccache_dev: oh oh, freeing busy cluster...\n");
      if(ccache[i].c_locked)printk("DMSDOS: free_ccache_dev: oh oh, freeing locked cluster...\n");
      FREE(ccache[i].c_data);
      ccache[i].c_flags=C_FREE;
      wake_up(&fullwait);
    }
  }
  unlock_ccache();
}

/* for memory management */
/* it is not clear whether locking is correct here -
   we simply ignore locked clusters like clusters that are currently in use
   (count!=0)
   it's also not clear whether we should lock the clusters before writing them
   there's a danger of deadlock since the whole cache is locked...
   so the clusters are not locked individually here...
   maybe we'd better not use ccache locking but cluster locking instead...
*/
void free_idle_ccache(void)
{ int i;
  struct super_block*sb;

  lock_ccache();
  for(i=0;i<CCACHESIZE;++i)
  { if(ccache[i].c_flags!=C_FREE&&ccache[i].c_count==0&&ccache[i].c_locked==0)
    { /* hmm... would be a good idea to make it depend on free system
         memory whether to discard which cached clusters... and how many...
         any ideas? */
      if(CURRENT_TIME-ccache[i].c_time>MAX_CCACHE_TIME)
      { /* throw away */
        /* but write it before if it is dirty... */
        if(ccache[i].c_flags==C_DIRTY_NORMAL)
        { sb=ccache[i].c_sb;
          if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
               ccache[i].c_clusternr,0,0)<0)continue;
        }
        if(ccache[i].c_flags==C_DIRTY_UNCOMPR)
        { sb=ccache[i].c_sb;
          if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
               ccache[i].c_clusternr,0,1)<0)continue;
        }
        FREE(ccache[i].c_data);
        ccache[i].c_flags=C_FREE;
        wake_up(&fullwait);
      }
    }
  }
  unlock_ccache();

  /* wake up forgotten fullwaits - for safety */
  wake_up(&fullwait);
}

/* for cluster cache syncing - writes all currently unused dirty clusters */
/* again it is not clear whether locking is correct here -
   we simply ignore locked clusters like clusters that are currently in use
   (count!=0)
   it's also not clear whether we should lock the clusters before writing them
   there's a danger of deadlock since the whole cache is locked...
   so the clusters are not locked individually here...
   maybe we'd better not use ccache locking but cluster locking instead...
*/
void sync_cluster_cache(int allow_daemon)
{ int i;
  struct super_block*sb;

  lock_ccache();
  
  for(i=0;i<CCACHESIZE;++i)
  { if(ccache[i].c_flags==C_DIRTY_NORMAL&&ccache[i].c_count==0&&ccache[i].c_locked==0)
    { sb=ccache[i].c_sb;
      if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                           ccache[i].c_clusternr,0,
                           allow_daemon?0:3)>=0)
        ccache[i].c_flags=C_USED;
    }
    if(ccache[i].c_flags==C_DIRTY_UNCOMPR&&ccache[i].c_count==0&&ccache[i].c_locked==0)
    { sb=ccache[i].c_sb;
      if(dmsdos_write_cluster(sb,ccache[i].c_data,ccache[i].c_length,
                           ccache[i].c_clusternr,0,1)>=0)
        ccache[i].c_flags=C_USED;
    }
  }
      
  unlock_ccache();
}

/* cluster delete - deletes this cluster from the cache and cancels ch_dirty */
void delete_cache_cluster(struct super_block*sb, int clusternr)
{ int i;

  retry:
  lock_ccache();
    
  for(i=0;i<CCACHESIZE;++i)
  { /* find it */
    if(ccache[i].c_clusternr==clusternr&&ccache[i].c_sb->s_dev==sb->s_dev
       &&ccache[i].c_flags!=C_FREE)
    { /* found */
      /* if it is locked we'd better wait here */
      if(ccache[i].c_locked)
      { unlock_ccache();
        sleep_on(&(ccache[i].c_wait));
        goto retry;
      }
      /* we don't lock the cluster since it is going to become free */
      /* is it currently in use ? */
      if(ccache[i].c_count)
      { ccache[i].c_clusternr=-1;
        /*ccache[i].c_sb=NULL;*/
        ccache[i].c_flags=C_DELETED;
      }
      else /* we can throw it away */
      { FREE(ccache[i].c_data);
        ccache[i].c_flags=C_FREE;
      }
      break; /* it shouldn't be in the cache twice :-) */
    }
  }

  /* inform the cluster handling routines so they can delete pointers in fs */
  dmsdos_write_cluster(sb,NULL,0,clusternr,0,0);
      
  unlock_ccache();
}

void log_ccache_statistics()
{ int j;
  int free;
  int used;
  int dirty_n;
  int dirty_u;
  int del;
  
  lock_ccache();
  
  printk("DMSDOS: ccache statistics:\n");
  
  free=0;
  used=0;
  dirty_n=0;
  dirty_u=0;
  del=0;
  
  for(j=0;j<CCACHESIZE;++j)
  { switch(ccache[j].c_flags)
    { case C_FREE: ++free; break;
      case C_USED: ++used; break;
      case C_DIRTY_NORMAL: ++dirty_n; break;
      case C_DIRTY_UNCOMPR: ++dirty_u; break;
      case C_DELETED: ++del; break;
      default: printk("DMSDOS: log_ccache_statistics: cannot happen.\n");
    }
  }
  
  printk("sum: free=%d used=%d dirty_n=%d dirty_u=%d del=%d\n",
         free,used,dirty_n,dirty_u,del);
         
  unlock_ccache();
}


/**********************************************************************/


struct buffer_head* dblspace_bread(struct super_block*sb,int vsector)
{ int dbl_clust;
  int sect_offs;
  struct buffer_head*bh;
  Cluster_head*ch;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
  
  if(vsector>=FAKED_ROOT_DIR_OFFSET&&
     vsector<FAKED_ROOT_DIR_OFFSET+dblsb->s_rootdiranzentry/16)
  { LOG_LLRW("DMSDOS: read_virtual_sector: rootdir sector_offset %d\n",
           vsector-FAKED_ROOT_DIR_OFFSET);
    return raw_bread(sb,dblsb->s_rootdir+vsector-FAKED_ROOT_DIR_OFFSET);
  }

  if(vsector<FAKED_DATA_START_OFFSET)  
  { err:
    printk("DMSDOS: illegal virtual sector %d, can't map to real sector\n",
           vsector);
    return NULL;
  }
  dbl_clust=((vsector-FAKED_DATA_START_OFFSET)/dblsb->s_sectperclust)+2;
  if(dbl_clust>dblsb->s_max_cluster)goto err;
  sect_offs=(vsector-FAKED_DATA_START_OFFSET)%dblsb->s_sectperclust;
  
  ch=ch_read(sb,dbl_clust);
  if(ch==NULL)
  { printk("DMSDOS: read_virtual_sector: read_cluster failed!\n");
    return NULL;
  }
  /* now setup a fake buffer_head */
  bh=MALLOC(sizeof(struct buffer_head));
  if(bh==NULL)
  { printk("DMSDOS: read_virtual_sector: no memory!\n");
    /*FREE(clusterd);*/
    ch_free(ch);
    return NULL;
  }
  bh->b_state=FAKED_BH_MAGIC;

  /* we must share the data */
  bh->b_data=&(ch->c_data[sect_offs*SECTOR_SIZE]);
  /* the cluster is *not* free yet, so DON'T call ch_free here */
  /* instead copy a pointer to the cluster_head somewhere so we can free it 
     later */
  bh->b_next=(struct buffer_head*)ch; /*the cast avoids compiler warning */
  return bh;
}

struct buffer_head *dblspace_getblk (
	struct super_block *sb,
	int block)
{       
        struct buffer_head*ret;
	ret=dblspace_bread(sb,block);
	if(ret==NULL)return ret;
	return ret;
}

void dblspace_brelse(struct super_block* sb,struct buffer_head*bh)
{ if(bh)
  { if(bh->b_state==FAKED_BH_MAGIC)
    { /* we free the cluster instead */
      /* a pointer to ch was saved in b_next */
      /* uh, we must check the size first.... uh no this is for bh_dirty.. */
      Cluster_head*ch=(Cluster_head*)bh->b_next;
      /*if((unsigned long)bh->b_data-(unsigned long)ch->c_data+SECTOR_SIZE>ch->c_length)
          ch->c_length=(unsigned long)bh->b_data-(unsigned long)ch->c_data+SECTOR_SIZE;*/
      ch_free(ch);
      FREE(bh);
      return;
    }
  }
  raw_brelse(sb,bh);
}


void dblspace_mark_buffer_dirty(struct super_block*sb,struct buffer_head*bh,
                                int dirty_val)
{ if(dirty_val==0)return;
  if(bh)
  { if(bh->b_state==FAKED_BH_MAGIC)
    { 
      /* a copy of ch was saved in b_next */
      Cluster_head*ch=(Cluster_head*)bh->b_next;
      struct super_block*sb=ch->c_sb;
      Dblsb*dblsb=MSDOS_SB(sb)->private_data;
      
      /* ensure cluster is marked large enough to hold the virtual sector */
      if((unsigned long)bh->b_data-(unsigned long)ch->c_data+SECTOR_SIZE>
                                                               ch->c_length)
          ch->c_length=(unsigned long)bh->b_data-(unsigned long)ch->c_data+
                                                                 SECTOR_SIZE;

      /* the virtual sector code is usually called by directory handling 
         routines, so we check whether we may compress a directory
         -- file_write calles ch_dirty directly */
      if(ch_dirty(ch,0,DIR_MAY_BE_COMPRESSED(dblsb)?0:1)<0)
      {
      /* now we have a serious problem here...
         ch_dirty failed... we are in danger of losing the data....
      */
         printk(KERN_EMERG "DMSDOS: Dirty virtual sector cannot be written - FILESYSTEM DAMAGE POSSIBLE! Trying to delay write.\n");
         ch_dirty_retry_until_success(ch,0,DIR_MAY_BE_COMPRESSED(dblsb)?0:1);
      }
      return; 
    }
  }

  raw_mark_buffer_dirty(sb,bh,1);
}

void dblspace_set_uptodate (
	struct super_block *sb,
	struct buffer_head *bh,
	int val)
{       
        if(bh->b_state==FAKED_BH_MAGIC)return; 
        raw_set_uptodate(sb,bh,val);
}

int dblspace_is_uptodate (
	struct super_block *sb,
	struct buffer_head *bh)
{
        if(bh->b_state==FAKED_BH_MAGIC)return 1;        
	return raw_is_uptodate(sb,bh);
}

void dblspace_ll_rw_block (
	struct super_block *sb,
	int opr,
	int nbreq,
	struct buffer_head *bh[32])
{
        /* code commented out because......
        
        int i;
        if(nbreq<=0||nbreq>32)panic("dblspace_ll_rw_block: illegal nbreq");
        if(opr==READA||opr==WRITEA)return; / we may ignore them /
        if(opr==WRITE)panic("dblspace_ll_rw_block: WRITE unsupported");
        if(opr!=READ)return;
        
        for(i=0;i<nbreq;++i)
        { bh[i]=dblspace_bread(sb,????
        }
        
        ......we just cannot do low-level access here */
        /* this might be wrong... */
        /* as far as I know, only the read_ahead code in file.c uses this */ 
        return;
}

#ifdef USE_READA_LIST
struct wait_queue * readawait=NULL;
int readalock=0;
void lock_reada(void)
{ while(readalock)sleep_on(&readawait);
  readalock=1;
}
void unlock_reada(void)
{ readalock=0;
  wake_up(&readawait);
}
struct
{ struct buffer_head*bh;
  struct super_block*sb;
} reada_list[READA_LIST_SIZE];
int ra_pos=0;

void init_reada_list(void)
{ int i;

  for(i=0;i<READA_LIST_SIZE;++i)
  { reada_list[i].bh=NULL;
    reada_list[i].sb=NULL;
  }
}

void kill_reada_list_dev(int dev)
{ int i;

  lock_reada();

  for(i=0;i<READA_LIST_SIZE;++i)
  { if(reada_list[i].bh)
    { if(reada_list[i].sb->s_dev==dev)
      { raw_brelse(reada_list[i].sb,reada_list[i].bh);
        reada_list[i].bh=NULL;
        reada_list[i].sb=NULL;
      }
    }
  }

  unlock_reada();
}

void stack_reada(struct super_block*sb,struct buffer_head*bh)
{
  lock_reada();

  if(reada_list[ra_pos].bh)
                   raw_brelse(reada_list[ra_pos].sb,reada_list[ra_pos].bh);

  reada_list[ra_pos].bh=bh;
  reada_list[ra_pos].sb=sb;

  ++ra_pos;
  if(ra_pos>=READA_LIST_SIZE)ra_pos=0;

  unlock_reada();
}
#endif /*USE_READA_LIST*/

void dblspace_reada(struct super_block*sb, int sector,int count)
{ 
  struct buffer_head*bhlist[MAX_READA];
  int i;
  struct buffer_head**bhlist2=bhlist;

  if(count>MAX_READA)count=MAX_READA;

  i=0;
  while(i<count)
  { 
    bhlist[i]=raw_getblk(sb,sector+i); /* get without read */
    if(bhlist[i]==NULL)break; /* failed, give up */
    if(raw_is_uptodate(sb,bhlist[i]))
    { raw_brelse(sb,bhlist[i]); /* according to breada the buffer must be
                                   freed here */
      break; /* has already been read, abort */
    }
    ++i;
  }

  count=i;

  if(count==0)return;

  /* uhh.. there's a hard limit of 32 in the fat filesystem... */
  if(count/32)
  { for(i=0;i<count/32;++i)
    { raw_ll_rw_block(sb,READA,32,bhlist2);
      bhlist2+=32;
    }
  }
  if(count%32)raw_ll_rw_block(sb,READA,count%32,bhlist2);
          /* place read command in list, but don't wait for it to finish */

  for(i=0;i<count;++i)
#ifdef USE_READA_LIST
            stack_reada(sb,bhlist[i]);
#else
            /* not a good idea - brelse calls wait_on_buffer....... */
            raw_brelse(sb,bhlist[i]);
#endif
}
